@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
GETBIAS:  Finds the selection bias for estimated wages, given a set of 
          model parameters
*/

proc(3)=getbias(hwage_fe, uwage_fe);

    local healsim, psim, nobs, maxage, simFEwage, hwage_sim, uwage_sim,  
          bvecgus, bvecbus, healbias, bvecgsm, bvecbsm, simFEwageg, simFEwageb, 
          grg, grb, rn;

    load path = ^shkpath healsim, psim;
    nobs   = rows(healsim);
    maxage = oldest(psim,(1-healsim));      @ get oldest age still working @
    clear healsim, psim;    
   "number of years, number of years in the lf= " lspan~maxage;?;

   @--------Estimate wage profiles, conditional on working and health---------@
    {simFEwage, hwage_sim, uwage_sim} = feunbal(nobs, maxage);

   @-------Compute the selection bias, using wage profiles from feunbal-------@
    {bvecgus, bvecbus, healbias} = compbias(hwage_fe, uwage_fe, simFEwage, maxage);

   @----------------------Smooth the estimated wage bias----------------------@
    simFEwage  = exp(simFEwage); 
    hwage_sim  = exp(hwage_sim); 
    uwage_sim  = exp(uwage_sim);
    bvecgsm    = polysm(bvecgus, ageseq, maxage); 
    bvecbsm    = polysm(bvecbus, ageseq, maxage);
    simFEwageg = simFEwage[1:maxage,.]; 
    simFEwageb = simFEwage[maxage+1:2*maxage,.];

    if lspan>maxage;
        bvecgsm = bvecgsm|zeros(lspan-maxage,1);
        bvecbsm = bvecbsm|zeros(lspan-maxage,1);
        grg     = hwage_sim[maxage+1:lspan]/hwage_sim[maxage]; @ total growth @
        grb     = uwage_sim[maxage+1:lspan]/uwage_sim[maxage];
        simFEwageg = simFEwageg|(simFEwageg[maxage]*grg); @ extrapolate with non-FE growth rates @        
        simFEwageb = simFEwageb|(simFEwageb[maxage]*grb);    
    endif;

    "Fixed-effect and non-fe wage profiles from simulations, healthy and unhealthy people";
    "        age         h-fe      h-non-fe         u-fe     u-non-fe";;
    ageseq~simFEwageg~hwage_sim~simFEwageb~uwage_sim;?;
    rn = maxage-1;
    "unsmoothed and smoothed wage growth bias profiles";
    "        age      h-usmthd      h-smthd     u-usmthd      u-smthd";;
    ageseq[2:rn+1,.]~bvecgus~bvecgsm[1:rn]~bvecbus~bvecbsm[1:rn];?;
    "underestimate (overestimate if negative) of the effect of health";
    "   on period-1 wages caused by selection = ";; healbias;?;

retp(bvecgsm, bvecbsm, healbias);
endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
OLDEST:  Finds the oldest age at which there is wage data for both healthy 
         and unhealthy people
*/
proc(1) = oldest(pmat,ghmat);
    local workgood, workbad, maxgood, maxbad, mymax;

    workgood = pmat.*ghmat;
    workbad  = pmat.*(1-ghmat);
    maxgood  = maxage(workgood);
    maxbad   = maxage(workbad);
    mymax    = minc(maxgood|maxbad);
retp(mymax); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
MAXAGE:  Finds the oldest age at which there is continuous wage data
*/

proc(1)=maxage(somemat);
    local meanmat, ii, j;

    meanmat=meanc(somemat);
    meanmat=meanmat./=0;

    j = indexcat(meanmat,0);
    if ismiss(j) == 1;
        j = lspan;
    else;
        j = minc(j);
        if j > 1;  j = j-1; endif;
    endif;

retp(j); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
COMPBIAS:  Compute extent to which fixed-effects estimator overestimates
           wage growth rates due to selection bias
*/
proc(3) = compbias(hwage_fe, uwage_fe, simFEwage, mxage);

    local biasvecg, biasvecb, estgrg, estgrb, meangrg, meangrb, healbias;

    hwage_fe = ln(hwage_fe[1:mxage,.]);
    uwage_fe = ln(uwage_fe[1:mxage,.]);
    estgrg   = simFEwage[2:mxage,.]-simFEwage[1:mxage-1,.];
    estgrb   = simFEwage[mxage+2:2*mxage,.]-simFEwage[mxage+1:(2*mxage)-1,.];
    meangrg  = hwage_fe[2:mxage,.]-hwage_fe[1:mxage-1,.];
    meangrb  = uwage_fe[2:mxage,.]-uwage_fe[1:mxage-1,.];

    biasvecg = meangrg-estgrg;
    biasvecb = meangrb-estgrb;
    healbias = (hwage_fe[1,.]-uwage_fe[1,.])-(simFEwage[1,.]-simFEwage[mxage+1,.]);

retp(biasvecg, biasvecb, healbias); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
POLYSM:  Smooth a function with a 5th-order polynomial.
         Lead with a zero and end with a zero to mitigate endpoint problems.
         The zeros are a good idea IF we believe that the function value is
         zero at the endpoints.
*/
proc(1)=polysm(somevec, agevec1, maxage);

    local agevec2, agevec3, agevec4, agevec5, localx, localb, smvec, 
          z1, z2, i;

    z1 = 1;  
    z2 = 2;
    somevec = zeros(z1,1)|somevec|zeros(z2,1);  @ somevec is initially (maxage-1)x1) @
    agevec1 = agevec1[1:maxage,.]|seqa(agevec1[maxage,.]+1,1,z2);
    agevec2 = agevec1.*agevec1;
    agevec3 = agevec2.*agevec1;
    agevec4 = agevec3.*agevec1;
    agevec5 = agevec4.*agevec1;
    localx  = ones(rows(agevec1),1)~agevec1~agevec2~agevec3~agevec4~agevec5;

    localb  = inv(localx'*localx);
    localb  = localb*localx'*somevec;
    smvec   = localb'*localx';
    smvec   = smvec';
    smvec   = smvec[1+z1:rows(somevec)-z2,.];

retp(smvec); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
FEUNBAL:  Find wage profile, conditional on working, using a fixed-effects
          estimator
*/
proc(3) = feunbal(nobs, mxage); 

    local wsim, healsim, psim, lnwmat, ghmat, pmat, hwmat, uwmat, healobs, 
          mpvec, obswmat, mwvec, wwvec, mwmat, liln, reps, i, xpx, xpy, 
          lxpx, lxpy, lwmat, lpmat, lghmat, hwage_sim, uwage_sim, simFEwage;
          
    load path = ^shkpath wsim, healsim, psim;
    lnwmat    = ln(wsim);                             @ do everything in logs @
    ghmat     = 1-healsim;                                  @ 1 -> bad health @
    pmat      = psim;
    clear wsim, healsim, psim;
    
   @------Compute mean wages by health status, NOT conditional on working-----@
    hwmat     = ghmat.*lnwmat; 
    uwmat     = (1-ghmat).*lnwmat;
    healobs   = meanc(ghmat);           @ Fraction of healthy obs at each age @
    hwage_sim = (meanc(hwmat))./healobs;
    uwage_sim = (meanc(uwmat))./(1-healobs);

   @-----------------Remove years where everyone has retired------------------@
    lnwmat    = lnwmat[.,1:mxage];
    pmat      = pmat[.,1:mxage];
    ghmat     = ghmat[.,1:mxage];
    
   @-------------De-mean (center) the wage data person-by-person--------------@
    mpvec     = sumc(pmat');  @ number of wage obs on each person, nx1 vector @
    obswmat   = lnwmat.*pmat;
    mwvec     = (sumc(obswmat'))./mpvec;  @ mean wages when working for each person @
    mwmat     = mwvec*ones(1,mxage);  @ Each person's lifetime mean for mxage years @
    mwmat     = mwmat.*pmat;
    lnwmat    = lnwmat.*pmat;
    lnwmat    = lnwmat-mwmat;              @ lnwmat demeaned person-by-person @

   @----Do a fixed-effect regression on wages for those in the labor force----@
   @------------------------Find X'X and X'y in steps-------------------------@
   @-------------This has to be done because of memory constraints------------@
    liln = nobs;
    reps = 1;
    if nobs >1000;
        if nobs le 10000;
            reps = 10;
            liln = nobs/10;
        else;
            reps = 100;
            liln = nobs/100;
        endif;
    endif;

    xpx = zeros(2*mxage-1,2*mxage-1);
    xpy = zeros(2*mxage-1,1);

    i=1; do while i le reps;
        lwmat  = lnwmat[(liln*(i-1))+1:liln*i,.];
        lpmat  = pmat[(liln*(i-1))+1:liln*i,.];
        lghmat = ghmat[(liln*(i-1))+1:liln*i,.];
        {lxpx, lxpy} = lfeunbal(lwmat,lpmat,lghmat,liln,mxage);
        xpx    = xpx+lxpx;
        xpy    = xpy+lxpy;
    i=i+1; endo;

   @--simFEwage gives us total wage growth, conditional on working and health--@
    simFEwage = inv(xpx)*xpy;

   @---Initialize profile with wages for healthy people in their first year---@
   @----------------Healthy and unhealthy profiles are stacked----------------@
    simFEwage = 0|simFEwage; 
    simFEwage = hwage_sim[1] + simFEwage;

retp(simFEwage, hwage_sim, uwage_sim); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
LFEUNBAL:  Returns X'X and X'y for subset of the data
*/
proc(2)=lfeunbal(lwmat, lpmat, lghmat, liln, mxage); 

    local xmat, ymat, oney, xchoiceg, xchoiceb, counter, i, j, lilcount,  
          meanx, lxpx, lxpy;

    xmat = zeros(liln*mxage,(mxage*2)-1);
    ymat = zeros(liln*mxage,1);
    oney = ones(2*mxage,1);

   @---Generate the "X" matrix, a matrix of age-health interaction dummies----@
    xchoiceg = eye(mxage)~zeros(mxage,mxage); 
    xchoiceg = xchoiceg[.,2:cols(xchoiceg)];           @ Dummies when healthy @
    xchoiceb = zeros(mxage,mxage)~eye(mxage); 
    xchoiceb = xchoiceb[.,2:cols(xchoiceb)];         @ Dummies when unhealthy @

    counter  = 1;
    i=1; do while i le liln;                            @ Loop through people @
        lilcount=1;
        j=1; do while j le mxage;                        @ Loop through years @
            if lpmat[i,j]==1;
                if lghmat[i,j]==1;
                    xmat[counter,.] = xchoiceg[j,.];
                else;
                    xmat[counter,.] = xchoiceb[j,.];
                endif;		
                ymat[counter,.] = lwmat[i,j];
                counter  = counter+1;
                lilcount = lilcount+1;
            endif;
        j=j+1; endo;

       @-------------------------Demean the X Matrix--------------------------@
        if lilcount>1;
            meanx = meanc(xmat[counter-lilcount+1:counter-1,.]);
            meanx = (meanx*ones(lilcount-1,1)')';
            xmat[counter-lilcount+1:counter-1,.]=
                xmat[counter-lilcount+1:counter-1,.]-meanx;
        endif;
    i=i+1; endo;

    xmat = xmat[1:counter-1,.]; 
    ymat = ymat[1:counter-1,.]; 

    lxpx = (xmat'*xmat);
    lxpy = (xmat'*ymat);

retp(lxpx, lxpy); endp;

@------------------------------------------------------------------------@
@------------------------------------------------------------------------@
/*
WAGEBIAS:  Solves the model, gets simulated data, and computes the
           incremental selection bias in wage profiles
*/
proc(3)=wagebias(params, biasvecg, biasvecb, healbias, hwage_fe, uwage_fe);

    local func, bvold, bvecg2, bvecb2, hlbias2, bv;

    func  = gmm(params); 
    bvold = biasvecb~biasvecg~(healbias|zeros(lspan-2,1));      
    {bvecg2, bvecb2, hlbias2} = getbias(hwage_fe, uwage_fe);
      
   @-------------Save total (not incremental) bias estimates-------------@
    biasvecg = biasvecg + bvecg2;
    biasvecb = biasvecb + bvecb2;
    healbias = healbias + hlbias2;
    bv       = biasvecb~biasvecg~(healbias|zeros(lspan-2,1));
    "Previous and revised estimates of wage selection bias ";
    "      bvecb0       bvecg0      hlbias0       bvecb1     bvecg1        hlbias1";;
    bvold~bv;?;
    save path = ^shkpath bv;

retp(biasvecg, biasvecb, healbias); endp; 

@------------------------------------------------------------------------@
@------------------------------------------------------------------------@

