/*
    Markch: Finds invariant distribution, conditional and unconditional
            expectations, unconditional variance and unconditional serial
            correlation for a finite state Markov Chain.

    John Jones:  14-Oct-1996
    Formulae from Hamilton, "Time Series Analysis" (1994)

@***************************************************************************@
    @***********************************************************************@
    proc chkchain(pmtx)
    -   Checks to make sure Markov chain is suitable for manipulation
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
        Output:
            isgood:     scalar (1=good,0=error)
    @***********************************************************************@
    proc chktval(pmtx,tval)
    -   Checks to make sure vector of values fits Markov chain
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
            tval:       kx1 vector of values for variable
        Output:
            isgood:     scalar (1=good,0=error)
    @***********************************************************************@
    proc fdchain(pmtx,tval)
    -   Converts Markov chain in levels to chain in first differences
        Inputs:
            pmtxl:      kxk transition matrix
                        rows denote time t values, cols time t+1 values
            tvall:      kx1 vector of values for variable
        Output:
            pmtxd:      (k^2)x(k^2) transition matrix of first differences
            tvald:      kx1 vector of first differences      
    @***********************************************************************@
    proc mrkinvar(pmtx)
    -   Finds the invariant matrix associated with a Markov Chain
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
        Output:
            invar:      1xk vector of invariant transition probabilities
                        (real portion)
    @***********************************************************************@
    proc mrkchpow(pmtx,power);
    -   Raises a Markov Chain to a Matrix power
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
            power:      power to which matrix is raised
        Output:
            chntopow:   (real portion of) pmtx raised to power
    @***********************************************************************@
    proc (2) = mrkchev(pmtx,tval);
    -   Finds conditional and unconditional expectations of Markov Chain
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
            tval:       kx1 vector of values for variable
        Output:
            condexp:    kx1 vector of conditional expectations
            ucondexp:   scalar of unconditional expectation
    @***********************************************************************@
    proc (2) = mrkchvar(pmtx,tval);
    -   Finds conditional and unconditional variance of Markov Chain
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
            tval:       kx1 vector of values for variable
        Output:
            condvar:    kx1 vector of conditional variances
            ucondvar:   scalar of unconditional variance
    @***********************************************************************@
    proc(2) =  mrkchcor(pmtx,tval);
    -   Finds 1st autovariance and serial correlation of Markov Chain
        Inputs:
            pmtx:       kxk transition matrix
                        rows denote time t values, cols time t+1 values
            tval:       kx1 vector of values for variable
        Output:
            autocov1:   scalar of 1st autocovariance
            serlcor:    scalar of serial correlation
    @***********************************************************************@
    proc(6) = markch(pmtx,tval);
    -   Markch: Finds invariant distribution, conditional and unconditional
        expectations, unconditional variance and unconditional serial
        correlation for a finite state Markov Chain.

        Inputs:
            pmtx:   kxk transition matrix
                    rows denote time t values, cols time t+1 values
            tval:   kx1 vector of values

        Outputs:
            invar:      kxk matrix with each row the invariant distribution
            condexp:    kx1 vector of conditional expectations
            ucondexp:   kx1 vector of unconditional expectations
            condvar:    kx1 vector of conditional variances
            ucondvar:   kxl of unconditional variances
            autocov:    scalar of 1st autocovariance
    @***********************************************************************@
@***************************************************************************@ */

proc chkchain(pmtx);
    local s, tots, isgood;
    isgood = 1;
    s = 1;
    s = rows(pmtx);

    if (s /= cols(pmtx));
        errorlog "Error:  Transition Matrix Not Square ";
        isgood = 0;
        goto isbad;
    endif;

    tots = sumc(pmtx');
    if (tots /= ones(s,1));
        errorlog "Error:  Transition Probabilities Don't Sum to 1";
        isgood = 0;
        goto isbad;
    endif;

    isbad:
    retp(isgood);
endp;

@***************************************************************************@

proc chktval(pmtx,tval);
    local s, tots, isgood;
    isgood = 1;
    s = 1;
    s = rows(pmtx);

    if (rows(tval) /= s);
        errorlog "Error: Transition Matrix Not Conforming to State Space";
        isgood = 0;
        goto isbad;
    endif;

    if (tval == tval[1,1]*ones(s,1));
        errorlog "Error:  All States Identical ";
        isgood = 0;
        goto isbad;
    endif;

    isbad:
    retp(isgood);
endp;

@***************************************************************************@

proc(2) = fdchain(pmtxl,tvall);

    local s,isgood,pmtxd,tvald,i,j,k,shft;

    s = rows(pmtxl);
    tvald = exp(-tvall)*exp(tvall'); @   v1-v1, v2-v1, ..., v1-v2, v2-v2,  @ 
    tvald = ln(vecr(tvald));
    pmtxd = zeros(s^2,s^2);

    isgood = chkchain(pmtxl);
    if (isgood == 0); goto done; endif;
    isgood = chktval(pmtxl,tvall);
    if (isgood == 0); goto done; endif;

    i=1; do until i>s;
        shft = (i-1)*s + 1;
        j=1; do until j>s;
            k = i + (j-1)*s;
            pmtxd[k,shft:i*s] = pmtxl[i,.];
        j=j+1; endo;
    i=i+1; endo;

    done:
    retp(pmtxd,tvald);
endp;

@***************************************************************************@

proc mrkinvar(pmtx);

    local s, isgood, invar, var, vai, ver, vei, va, ve, A;

    s = rows(pmtx);
    invar = -ones(s,1);

    isgood = chkchain(pmtx);
    if (isgood == 0); goto done; endif;

    {var,vai,ver,vei} = eigrg2(pmtx');  @  Note pmtx is transposed @
    va = complex(var,vai);      @  eigenvalues     @
    ve = complex(ver,vei);      @  eigenvectors    @

@   round to 12 degrees of precision    @
@   va = round(va*1e12)/1e12;           @

    va = (va.== ones(s,1));

    if (sumc(va) > 1);
        errorlog "Error:  Multiple Invariant Measures--Can't Help You";
        goto done;
    endif;

    A = (eye(s)-pmtx')|ones(1,s);       @  See Hamilton, p. 684 @
    A = invpd(A'A)*A';
    invar = real(A[.,s+1]);

    done:
    retp(invar');
endp;

@***************************************************************************@

proc mrkchpow(pmtx,power);

    local s, isgood, var, vai, ver, vei, va, ve, vadiag, chntopow;

    s = 1;
    s = rows(pmtx);
    chntopow = -ones(s,s);

    isgood = chkchain(pmtx);
    if (isgood == 0); goto done; endif;

    {var,vai,ver,vei} = eigrg2(pmtx');  @  Note pmtx is transposed @
    va = complex(var,vai);      @  eigenvalues     @
    ve = complex(ver,vei);      @  eigenvectors    @

@   round to 12 degrees of precision @
@   va = round(va*1e12)/1e12;        @

    va = va.^power;
    vadiag = diagrv(eye(s),va);
    chntopow = ve*vadiag*inv(ve);
    chntopow = real(chntopow');
    done:
    retp(chntopow);
endp;

@***************************************************************************@

proc(2) =  mrkchev(pmtx,tval);

    local s, isgood, invar, condexp, ucondexp;

    s = 1;
    s = rows(pmtx);
    condexp = -1e7*ones(s,1);
    ucondexp = -1e7;

    isgood = chkchain(pmtx);
    if (isgood == 0); goto done; endif;
    isgood = chktval(pmtx,tval);
    if (isgood == 0); goto done; endif;

    invar = mrkinvar(pmtx);
    if (invar==-ones(1,s)); goto done; endif;
    condexp = pmtx*tval;
    ucondexp = invar*tval;
    done:
    retp(condexp,ucondexp);
endp;

@***************************************************************************@

proc(2) =  mrkchvar(pmtx,tval);

    local s, isgood, invar, condexp, ucondexp, condev, condevsq, ucondev,
            condvar, ucondvar, j;

    s = 1;
    s = rows(pmtx);
    condexp = zeros(s,1);
    ucondexp = 0;
    condvar = -ones(s,1);
    ucondvar = -1;

    isgood = chkchain(pmtx);
    if (isgood == 0); goto done; endif;
    isgood = chktval(pmtx,tval);
    if (isgood == 0); goto done; endif;

    invar = mrkinvar(pmtx);
    if (invar==-ones(1,s)); goto done; endif;
    condexp = pmtx*tval;
    ucondexp = invar*tval;

/*  Conditional Deviations:  Each column j gives realizations
    of tval at time t+1 less the expectation of tval given state
    j at time t */

    condev = ones(s,1)*tval' - condexp*ones(1,s);
    condev = condev';
    condevsq = condev.^2;

    j=1; do until j>s;
        condvar[j,1] = pmtx[j,.]*condevsq[.,j];
        j=j+1;
    endo;
    condvar = real(condvar);

    ucondev = tval - ucondexp*ones(s,1);
    ucondvar = real(invar*(ucondev.^2));
    done:
    retp(condvar,ucondvar);

endp;

@***************************************************************************@

proc(2) =  mrkchcor(pmtx,tval);

    local s, isgood, invar, condexp, ucondexp, ucondev, ucondvar,
            autocov1, serlcor;

    s = 1;
    s = rows(pmtx);
    ucondexp = 0;
    ucondvar = -1;
    autocov1 = 0;
    serlcor = -99;

    isgood = chkchain(pmtx);
    if (isgood == 0); goto done; endif;
    isgood = chktval(pmtx,tval);
    if (isgood == 0); goto done; endif;

    invar = mrkinvar(pmtx);
    if (invar==-ones(1,s)); goto done; endif;
    ucondexp = invar*tval;

    ucondev = tval - ucondexp*ones(s,1);
    ucondvar = real(invar*(ucondev.^2));

    autocov1 = (invar'.*ucondev)'*pmtx*ucondev;
    serlcor = autocov1/ucondvar;
    done:
    retp(autocov1,serlcor);
endp;

@***************************************************************************@

proc(6) = markch(pmtx,tval);

    local s, invar, condexp, ucondexp, condvar, ucondvar, serlcor, condev, 
          condevsq, ucondev, j, autocov1, isgood;

    s = rows(pmtx);

    invar = zeros(s,s);
    condexp = ones(s,1);
    ucondexp = ones(s,1);
    condvar = zeros(s,1);
    condev = zeros(s,s);
    ucondvar = zeros(s,1);
    serlcor = 0;

    isgood = chkchain(pmtx);
    if (isgood == 0); goto done; endif;
    isgood = chktval(pmtx,tval);
    if (isgood == 0); goto done; endif;

    invar = mrkinvar(pmtx);
    if (invar==-ones(1,s)); goto done; endif;

    condexp = pmtx*tval;
    ucondexp = invar*tval;

@   Conditional Deviations:  Each column j gives realizations of tval at  @
@   time t+1 less the expectation of tval given state j at time t         @

    condev = ones(s,1)*tval' - condexp*ones(1,s);
    condev = condev';
    condevsq = condev.^2;

    j=1; do until j>s;
        condvar[j,1] = pmtx[j,.]*condevsq[.,j];
        j=j+1;
    endo;
    condvar = real(condvar);

    ucondev = tval - ucondexp;
    ucondvar = invar[1,.]*real((ucondev.^2));

    autocov1 = ((invar[1,.]').*ucondev)'*pmtx*ucondev;
    serlcor = autocov1/ucondvar[1,1];
    "Possible Values:  "; tval';"";
    "Transition Matrix [i,j] (i->t,j->t+1):  "; pmtx;"";
    "Invariant Distribution:  "; invar[1,.];"";
    "Conditional Expectations:  "; condexp';
    "Conditional Variance:  "; condvar';
    "Unconditional Expectations:  ";; ucondexp[1,1];
    "Unconditional Variance:      ";; ucondvar[1,1];
    "Unconditional Std. Deviation:";; sqrt(ucondvar[1,1]);
    "1st Autocovariance:          ";; autocov1;
    "Serial Correlation:          ";; serlcor;"";

    done:
    retp(invar,condexp,ucondexp,condvar,ucondvar,autocov1);
endp;